import datetime
import json

import numpy as np
import pytest
from bson.objectid import ObjectId
from monty.json import MSONable

from emmet.core.utils import DocEnum, ValueEnum, jsanitize


def test_jsanitize():
    """
    Tests emmet Jsanitize which converts MSONable classes into dicts
    """
    # clean_json should have no effect on None types.
    d = {"hello": 1, "world": None}
    clean = jsanitize(d)
    assert clean["world"] is None
    assert json.loads(json.dumps(d)) == json.loads(json.dumps(clean))

    d = {"hello": GoodMSONClass(1, 2, 3)}
    with pytest.raises(TypeError):
        json.dumps(d)

    clean = jsanitize(d)
    assert isinstance(clean["hello"], dict)
    clean_strict = jsanitize(d, strict=True)
    assert clean_strict["hello"]["a"] == 1
    assert clean_strict["hello"]["b"] == 2

    d = {"dt": datetime.datetime.now()}
    clean = jsanitize(d)
    assert isinstance(clean["dt"], str)
    clean = jsanitize(d, allow_bson=True)
    assert isinstance(clean["dt"], datetime.datetime)

    d = {
        "a": ["b", np.array([1, 2, 3])],
        "b": ObjectId.from_datetime(datetime.datetime.now()),
    }
    clean = jsanitize(d)
    assert clean["a"] == ["b", [1, 2, 3]]
    assert isinstance(clean["b"], str)

    rnd_bin = bytes(np.random.rand(10))
    d = {"a": bytes(rnd_bin)}
    clean = jsanitize(d, allow_bson=True)
    assert clean["a"] == bytes(rnd_bin)
    assert isinstance(clean["a"], bytes)


class GoodMSONClass(MSONable):
    def __init__(self, a, b, c, d=1, **kwargs):
        self.a = a
        self.b = b
        self._c = c
        self._d = d
        self.kwargs = kwargs

    def __eq__(self, other):
        return (
            self.a == other.a
            and self.b == other.b
            and self._c == other._c
            and self._d == other._d
            and self.kwargs == other.kwargs
        )


def test_value_enum():
    class TempEnum(ValueEnum):
        A = "A"
        B = "B"

    assert str(TempEnum.A) == "A"
    assert str(TempEnum.B) == "B"


def test_doc_enum():
    class TestEnum(DocEnum):
        A = "A", "Describes A"
        B = "B", "Might describe B"

    assert str(TestEnum.A) == "A"
    assert TestEnum.B.__doc__ == "Might describe B"



from typing import Dict

import numpy as np
from pydantic import Field, root_validator
from pymatgen.analysis.diffraction.xrd import (
    WAVELENGTHS,
    DiffractionPattern,
    XRDCalculator,
)
from pymatgen.core import Structure
from pymatgen.core.periodic_table import Element

from emmet.core.mpid import MPID
from emmet.core.spectrum import SpectrumDoc
from emmet.core.utils import ValueEnum


class Edge(ValueEnum):
    K_Alpha = "Ka"
    K_Alpha1 = "Ka1"
    K_Alpha2 = "Ka2"
    K_Beta = "Kb"
    K_Beta1 = "Kb1"
    K_Beta2 = "Kb2"


class XRDDoc(SpectrumDoc):
    """
    Document describing a XRD Diffraction Pattern
    """

    spectrum_name = "XRD"

    spectrum: DiffractionPattern
    min_two_theta: float
    max_two_theta: float
    wavelength: float = Field(..., description="Wavelength for the diffraction source.")
    target: Element = Field(
        None, description="Target element for the diffraction source."
    )
    edge: Edge = Field(None, description="Atomic edge for the diffraction source.")

    @root_validator(pre=True)
    def get_target_and_edge(cls, values: Dict):
        print("Validations")
        # Only do this if neither target not edge is defined
        if "target" not in values and "edge" not in values:
            try:
                pymatgen_wavelength = next(
                    k
                    for k, v in WAVELENGTHS.items()
                    if np.allclose(values["wavelength"], v)
                )
                values["target"] = pymatgen_wavelength[:2]
                values["edge"] = pymatgen_wavelength[2:]

            except Exception:
                return values
        return values

    @classmethod
    def from_structure(  # type: ignore[override]
        cls,
        material_id: MPID,
        spectrum_id: str,
        structure: Structure,
        wavelength: float,
        min_two_theta=0,
        max_two_theta=180,
        symprec=0.1,
        **kwargs,
    ) -> "XRDDoc":
        calc = XRDCalculator(wavelength=wavelength, symprec=symprec)
        pattern = calc.get_pattern(
            structure, two_theta_range=(min_two_theta, max_two_theta)
        )

        return super().from_structure(
            material_id=material_id,
            spectrum_id=spectrum_id,
            meta_structure=structure,
            spectrum=pattern,
            wavelength=wavelength,
            min_two_theta=min_two_theta,
            max_two_theta=max_two_theta,
            **kwargs,
        )

    @classmethod
    def from_target(
        cls,
        material_id: MPID,
        structure: Structure,
        target: Element,
        edge: Edge,
        min_two_theta=0,
        max_two_theta=180,
        symprec=0.1,
        **kwargs,
    ) -> "XRDDoc":
        if f"{target}{edge}" not in WAVELENGTHS:
            raise ValueError(f"{target}{edge} not in pymatgen wavelenghts dictionarty")

        wavelength = WAVELENGTHS[f"{target}{edge}"]
        spectrum_id = f"{material_id}-{target}{edge}"

        return cls.from_structure(
            material_id=material_id,
            spectrum_id=spectrum_id,
            structure=structure,
            wavelength=wavelength,
            target=target,
            edge=edge,
            min_two_theta=min_two_theta,
            max_two_theta=max_two_theta,
            **kwargs,
        )



import json
import datetime
import copy
import pickle

import pytest

from monty.io import zopen
from monty.serialization import loadfn

from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.molecules.redox import RedoxDoc


@pytest.fixture(scope="session")
def test_mols(test_dir):
    mols = loadfn((test_dir / "builder_mols_set.json").as_posix())
    mols = [MoleculeDoc(**x) for x in mols]

    return mols


@pytest.mark.skip(reason="Waiting on molecule update.")
def test_redox(test_mols):
    entries = list()
    for m in test_mols:
        for e in m.entries:
            entries.append(e)

    redox_docs = RedoxDoc.from_entries(entries)

    assert len(redox_docs) == 18
    assert redox_docs[0].property_name == "redox"

    assert redox_docs[1].electron_affinity == pytest.approx(-1.888795786559726)
    assert str(redox_docs[1].ea_id) == "116081"
    assert redox_docs[1].ionization_energy == pytest.approx(2.392317962137756)
    assert str(redox_docs[1].ie_id) == "116001"



""" Core definition of a Task Document which represents a calculation from some program"""
from datetime import datetime

from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.mpid import MPID


class TaskDocument(EmmetBaseModel):
    """
    Definition of Task Document
    """

    calc_code: str = Field(description="The calculation code used to compute this task")
    version: str = Field(None, description="The version of the calculation code")
    dir_name: str = Field(None, description="The directory for this task")
    task_id: MPID = Field(None, description="the Task ID For this document")

    completed: bool = Field(False, description="Whether this calculation completed")
    completed_at: datetime = Field(None, description="Timestamp for when this task was completed")
    last_updated: datetime = Field(
        default_factory=datetime.utcnow,
        description="Timestamp for this task document was last updated",
    )



import pytest
from pymatgen.core import Structure
from pymatgen.io.vasp.inputs import Kpoints 
from monty.serialization import loadfn
from emmet.core.absorption import AbsorptionDoc
from emmet.core.utils import jsanitize
import numpy as np 

@pytest.fixture(scope="session")
def absorption_test_data(test_dir):
    return loadfn(test_dir / "sample_absorptions.json")


def test_absorption_doc(absorption_test_data):
    absorption_coeff = np.array([
        0,
        5504161142.509,
        16485480924.4252,
        41235259342.4927,
        76990619286.8861,
        109929386572.273,
        164921201202.527,
        230913400825.579,
        285790460873.292,
        371002598552.062
    ])

    imag_dielectric = [
        0.0,
        0.0002,
        0.0003
    ]

    energies = [
        0,
        0.0309,
        0.0617,
        0.0926,
        0.1235,
        0.1543,
        0.1852,
        0.2161,
        0.2469,
        0.2778
    ]

    data = absorption_test_data
    structure = Structure.from_dict(jsanitize(data["input"]["structure"]))
    task_id = data["task_id"]
    kpoints = data["orig_inputs"]["kpoints"]

    doc = AbsorptionDoc.from_structure(
        structure=structure,
        material_id = 'mp-{}'.format(task_id),
        task_id=task_id,
        deprecated=False,
        energies=data["output"]["dielectric"]["energy"],
        real_d=data["output"]["dielectric"]["real"],
        imag_d=data["output"]["dielectric"]["imag"],
        absorption_co=data["output"]["optical_absorption_coeff"],
        bandgap=data["output"]["bandgap"],
        nkpoints=kpoints.num_kpts,
        is_hubbard=False,
    )

    assert doc is not None
    assert doc.property_name == "Optical absorption spectrum"
    assert doc.energies[0:10] == energies
    assert doc.material_id == 'mp-1316'
    assert doc.absorption_coefficient[0:10] == list(absorption_coeff)
    assert doc.average_imaginary_dielectric[0:3] == imag_dielectric
    assert doc.bandgap == 4.4652



from datetime import datetime
from typing import List, Union, Dict, Tuple, Sequence

from pydantic import Field
import numpy as np
from emmet.core.base import EmmetBaseModel
from pymatgen.core import Structure
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry

try:
    from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
    from pymatgen.analysis.diffusion.utils.supercells import get_sc_fromstruct
except ImportError:
    raise ImportError("Install pymatgen-analysis-diffusion to use MigrationGraphDoc")


class MigrationGraphDoc(EmmetBaseModel):
    """
    MigrationGraph Doc.
    Stores MigrationGraph and info such as ComputedStructureEntries (ComputedEntry can be used for working ion)
    and cutoff distance that are used to generated the object.

    Note: this doc is not self-contained within pymatgen, as it has dependence on pymatgen.analysis.diffusion,
    a namespace package aka pymatgen-diffusion.
    """

    battery_id: str = Field(..., description="The battery id for this MigrationGraphDoc")

    last_updated: datetime = Field(
        None, description="Timestamp for the most recent calculation for this MigrationGraph document.",
    )

    warnings: Sequence[str] = Field([], description="Any warnings related to this property.")

    deprecated: bool = Field(
        False,
        description="Indicates whether a migration graph fails to be constructed from the provided entries. Defaults to False, indicating mg can be constructed from entries.",  # noqa: E501
    )

    hop_cutoff: float = Field(
        None, description="The numerical value in angstroms used to cap the maximum length of a hop.",
    )

    entries_for_generation: List[ComputedStructureEntry] = Field(
        None,
        description="A list of ComputedStructureEntries used to generate the structure with all working ion sites.",
    )

    working_ion_entry: Union[ComputedEntry, ComputedStructureEntry] = Field(
        None, description="The ComputedStructureEntry of the working ion."
    )

    migration_graph: MigrationGraph = Field(
        None, description="The MigrationGraph object as defined in pymatgen.analysis.diffusion.",
    )

    populate_sc_fields: bool = Field(
        True, description="Flag indicating whether this document has populated the supercell fields",
    )

    min_length_sc: float = Field(
        None, description="The minimum length used to generate supercell using pymatgen.",
    )

    minmax_num_atoms: Tuple[int, int] = Field(
        None, description="The min/max number of atoms used to genreate supercell using pymatgen.",
    )

    matrix_supercell_structure: Structure = Field(
        None,
        description="The matrix suprcell structure that does not contain the mobile ions for the purpose of migration analysis.",  # noqa: E501
    )

    conversion_matrix: List[List[Union[int, float]]] = Field(
        None, description="The conversion matrix used to convert unit cell to supercell.",
    )

    inserted_ion_coords: List[Dict[str, Union[List[float], str, int]]] = Field(
        None, description="A dictionary containing all mobile ion fractional coordinates in terms of supercell.",
    )

    insert_coords_combo: List[str] = Field(
        None,
        description="A list of combinations 'a+b' to designate hops in the supercell. Each combo should correspond to one unique hop in MigrationGraph.",  # noqa: E501
    )

    @classmethod
    def from_entries_and_distance(
        cls,
        battery_id: str,
        grouped_entries: List[ComputedStructureEntry],
        working_ion_entry: Union[ComputedEntry, ComputedStructureEntry],
        hop_cutoff: float,
        populate_sc_fields: bool = True,
        ltol: float = 0.2,
        stol: float = 0.3,
        angle_tol: float = 5,
        **kwargs,
    ) -> Union["MigrationGraphDoc", None]:
        """
        This classmethod takes a group of ComputedStructureEntries (can also use ComputedEntry for wi) and generates
        a full sites structure.
        Then a MigrationGraph object is generated with with_distance() method with a designated cutoff.
        If populate_sc_fields set to True, this method will populate the supercell related fields. Required kwargs are
        min_length_sc and minmax_num_atoms.
        """

        ranked_structures = MigrationGraph.get_structure_from_entries(
            entries=grouped_entries, migrating_ion_entry=working_ion_entry
        )
        max_sites_struct = ranked_structures[0]

        migration_graph = MigrationGraph.with_distance(
            structure=max_sites_struct,
            migrating_specie=working_ion_entry.composition.chemical_system,
            max_distance=hop_cutoff,
        )

        if not populate_sc_fields:
            return cls(
                battery_id=battery_id,
                hop_cutoff=hop_cutoff,
                entries_for_generation=grouped_entries,
                working_ion_entry=working_ion_entry,
                migration_graph=migration_graph,
                **kwargs,
            )

        else:

            if all(arg in kwargs for arg in ["min_length_sc", "minmax_num_atoms"]):
                sm = StructureMatcher(ltol, stol, angle_tol)
                (
                    host_sc,
                    sc_mat,
                    min_length_sc,
                    minmax_num_atoms,
                    coords_list,
                    combo,
                ) = MigrationGraphDoc.generate_sc_fields(
                    mg=migration_graph,
                    min_length_sc=kwargs["min_length_sc"],
                    minmax_num_atoms=kwargs["minmax_num_atoms"],
                    sm=sm,
                )

                return cls(
                    battery_id=battery_id,
                    hop_cutoff=hop_cutoff,
                    entries_for_generation=grouped_entries,
                    working_ion_entry=working_ion_entry,
                    migration_graph=migration_graph,
                    matrix_supercell_structure=host_sc,
                    conversion_matrix=sc_mat,
                    inserted_ion_coords=coords_list,
                    insert_coords_combo=combo,
                    **kwargs,
                )

            else:
                raise TypeError(
                    "Please make sure to have kwargs min_length_sc and minmax_num_atoms if populate_sc_fields is set to True."  # noqa: E501
                )

    @staticmethod
    def generate_sc_fields(
        mg: MigrationGraph, min_length_sc: float, minmax_num_atoms: Tuple[int, int], sm: StructureMatcher,
    ):
        min_length_sc = min_length_sc
        minmax_num_atoms = minmax_num_atoms

        sc_mat = get_sc_fromstruct(
            base_struct=mg.structure,
            min_atoms=minmax_num_atoms[0],
            max_atoms=minmax_num_atoms[1],
            min_length=min_length_sc,
        )

        sc_mat = sc_mat.tolist()
        host_sc = mg.host_structure * sc_mat
        working_ion = mg.only_sites[0].species_string

        coords_list = MigrationGraphDoc.ordered_sc_site_list(mg.only_sites, sc_mat)
        combo, coords_list = MigrationGraphDoc.get_hop_sc_combo(
            mg.unique_hops, sc_mat, sm, host_sc, working_ion, coords_list
        )

        return host_sc, sc_mat, min_length_sc, minmax_num_atoms, coords_list, combo

    @staticmethod
    def ordered_sc_site_list(uc_sites_only: Structure, sc_mat: List[List[Union[int, float]]]):
        uc_no_site = uc_sites_only.copy()
        uc_no_site.remove_sites(range(len(uc_sites_only)))
        working_ion = uc_sites_only[0].species_string
        sc_site_dict = {}  # type: dict

        for i, e in enumerate(uc_sites_only):
            uc_one_set = uc_no_site.copy()
            uc_one_set.insert(0, working_ion, e.frac_coords)
            sc_one_set = uc_one_set * sc_mat
            for index in range(len(sc_one_set)):
                sc_site_dict[len(sc_site_dict) + 1] = {
                    "uc_site_type": i,
                    "site_frac_coords": list(sc_one_set[index].frac_coords),
                    # "extra_site": False
                }

        ordered_site_list = [
            e
            for i, e in enumerate(
                sorted(sc_site_dict.values(), key=lambda v: float(np.linalg.norm(v["site_frac_coords"])),)
            )
        ]
        return ordered_site_list

    @staticmethod
    def get_hop_sc_combo(
        unique_hops: Dict,
        sc_mat: List[List[Union[int, float]]],
        sm: StructureMatcher,
        host_sc: Structure,
        working_ion: str,
        ordered_sc_site_list: list,
    ):
        combo = []

        unique_hops = {k: v for k, v in sorted(unique_hops.items())}
        for one_hop in unique_hops.values():
            added = False
            sc_isite_set = {k: v for k, v in enumerate(ordered_sc_site_list) if v["uc_site_type"] == one_hop["iindex"]}
            sc_esite_set = {k: v for k, v in enumerate(ordered_sc_site_list) if v["uc_site_type"] == one_hop["eindex"]}
            for sc_iindex, sc_isite in sc_isite_set.items():
                for sc_eindex, sc_esite in sc_esite_set.items():
                    sc_check = host_sc.copy()
                    sc_check.insert(0, working_ion, sc_isite["site_frac_coords"])
                    sc_check.insert(1, working_ion, sc_esite["site_frac_coords"])
                    if MigrationGraphDoc.compare_sc_one_hop(
                        one_hop,
                        sc_mat,
                        sm,
                        host_sc,
                        sc_check,
                        working_ion,
                        (sc_isite["uc_site_type"], sc_esite["uc_site_type"]),
                    ):
                        combo.append(f"{sc_iindex}+{sc_eindex}")
                        added = True
                        break
                if added:
                    break

            if not added:
                new_combo, ordered_sc_site_list = MigrationGraphDoc.append_new_site(
                    host_sc, ordered_sc_site_list, one_hop, sc_mat, working_ion
                )
                combo.append(new_combo)

        return combo, ordered_sc_site_list

    @staticmethod
    def compare_sc_one_hop(
        one_hop: Dict,
        sc_mat: List,
        sm: StructureMatcher,
        host_sc: Structure,
        sc_check: Structure,
        working_ion: str,
        uc_site_types: Tuple[int, int],
    ):
        sc_mat_inv = np.linalg.inv(sc_mat)
        convert_sc_icoords = np.dot(one_hop["ipos"], sc_mat_inv)
        convert_sc_ecoords = np.dot(one_hop["epos"], sc_mat_inv)
        convert_sc = host_sc.copy()
        convert_sc.insert(0, working_ion, convert_sc_icoords)
        convert_sc.insert(1, working_ion, convert_sc_ecoords)

        if sm.fit(convert_sc, sc_check):
            one_hop_dis = one_hop["hop"].length
            sc_check_hop_dis = np.linalg.norm(sc_check[0].coords - sc_check[1].coords)
            if np.isclose(one_hop_dis, sc_check_hop_dis, atol=1e-5):
                if (
                    one_hop["iindex"] == uc_site_types[0]
                    and one_hop["eindex"] == uc_site_types[1]
                ):
                    return True

        return False

    @staticmethod
    def append_new_site(
        host_sc: Structure,
        ordered_sc_site_list: list,
        one_hop: Dict,
        sc_mat: List[List[Union[int, float]]],
        working_ion: str,
    ):
        sc_mat_inv = np.linalg.inv(sc_mat)
        sc_ipos = np.dot(one_hop["ipos"], sc_mat_inv)
        sc_epos = np.dot(one_hop["epos"], sc_mat_inv)
        sc_iindex, sc_eindex = None, None
        host_sc_insert = host_sc.copy()

        for k, v in enumerate(ordered_sc_site_list):
            if np.allclose(sc_ipos, v["site_frac_coords"], rtol=0.1, atol=0.1):
                sc_iindex = k
            if np.allclose(sc_epos, v["site_frac_coords"], rtol=0.1, atol=0.1):
                sc_eindex = k

        if sc_iindex is None:
            host_sc_insert.insert(0, working_ion, sc_ipos)
            ordered_sc_site_list.append(
                {
                    "uc_site_type": one_hop["iindex"],
                    "site_frac_coords": list(host_sc_insert[0].frac_coords),
                    "extra_site": True,
                }
            )
            sc_iindex = len(ordered_sc_site_list) - 1
        if sc_eindex is None:
            host_sc_insert.insert(0, working_ion, sc_epos)
            ordered_sc_site_list.append(
                {
                    "uc_site_type": one_hop["eindex"],
                    "site_frac_coords": list(host_sc_insert[0].frac_coords),
                    "extra_site": True,
                }
            )
            sc_eindex = len(ordered_sc_site_list) - 1

        return f"{sc_iindex}+{sc_eindex}", ordered_sc_site_list

    def get_distinct_hop_sites(self) -> Tuple[List, List[str], Dict]:
        """
        This is a utils function that converts the site dict and combo into a site list and combo that contain only distince endpoints used the combos. # noqa: E501
        """
        if self.inserted_ion_coords is None or self.insert_coords_combo is None:
            raise TypeError(
                "Please make sure that the MGDoc passed in has inserted_ion_coords and inserted_coords_combo fields filled."  # noqa: E501
            )

        else:
            dis_sites_list = []
            dis_combo_list = []
            mgdoc_sites_mapping = {}  # type: dict
            combo_mapping = {}

            for one_combo in self.insert_coords_combo:
                ini, end = list(map(int, one_combo.split("+")))

                if ini in mgdoc_sites_mapping.keys():
                    dis_ini = mgdoc_sites_mapping[ini]
                else:
                    dis_sites_list.append(
                        list(self.inserted_ion_coords[ini]["site_frac_coords"])  # type: ignore
                    )
                    dis_ini = len(dis_sites_list) - 1
                    mgdoc_sites_mapping[ini] = dis_ini
                if end in mgdoc_sites_mapping.keys():
                    dis_end = mgdoc_sites_mapping[end]
                else:
                    dis_sites_list.append(
                        list(self.inserted_ion_coords[end]["site_frac_coords"])  # type: ignore
                    )
                    dis_end = len(dis_sites_list) - 1
                    mgdoc_sites_mapping[end] = dis_end

                dis_combo = f"{dis_ini}+{dis_end}"
                dis_combo_list.append(dis_combo)
                combo_mapping[dis_combo] = one_combo

            return dis_sites_list, dis_combo_list, combo_mapping



from typing import List

from pydantic import Field

from pymatgen.core.structure import Molecule

from emmet.core.mpid import MPID
from emmet.core.material import PropertyOrigin
from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.molecule_property import PropertyDoc


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class VibrationDoc(PropertyDoc):

    property_name = "vibrations"

    molecule: Molecule = Field(..., description="Molecular structure")

    frequencies: List[float] = Field(
        ..., description="List of molecular vibrational frequencies"
    )

    frequency_modes: List[List[List[float]]] = Field(
        ..., description="Vibrational frequency modes of the molecule"
    )

    ir_intensities: List[float] = Field(
        ...,
        title="IR intensities",
        description="Intensities for IR vibrational spectrum peaks",
    )

    ir_activities: List = Field(
        ...,
        title="IR activities",
        description="List indicating if frequency-modes are IR-active",
    )

    @classmethod
    def from_task(
        cls, task: TaskDocument, molecule_id: MPID, deprecated: bool = False, **kwargs
    ):  # type: ignore[override]
        """
        Construct a vibration document from a task document

        :param task: document from which vibrational properties can be extracted
        :param molecule_id: mpid
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.output.frequencies is None:
            raise Exception("No frequencies in task!")

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        frequencies = task.output.frequencies
        frequency_modes = None
        intensities = None
        active = None
        for calc in task.calcs_reversed:
            if (
                calc.get("frequency_mode_vectors", None) is not None
                and frequency_modes is None
            ):
                frequency_modes = calc.get("frequency_mode_vectors")

            if calc.get("IR_intens", None) is not None and intensities is None:
                intensities = calc.get("IR_intens")

            if calc.get("IR_active", None) is not None and active is None:
                active = calc.get("IR_active")

            if all([x is not None for x in [frequency_modes, intensities, active]]):
                break

        if frequency_modes is None:
            raise Exception("No frequency modes in task!")
        elif intensities is None:
            raise Exception("No IR intensities in task!")
        elif active is None:
            raise Exception("No IR activities in task!")

        warnings = list()
        if frequencies[0] < 0.0:
            warnings.append("Imaginary frequencies")

        return super().from_molecule(
            meta_molecule=mol,
            molecule_id=molecule_id,
            molecule=mol,
            frequencies=frequencies,
            frequency_modes=frequency_modes,
            ir_intensities=intensities,
            ir_activities=active,
            origins=[PropertyOrigin(name="vibrations", task_id=task.task_id)],
            deprecated=deprecated,
            warnings=warnings,
            **kwargs
        )



import json
import datetime
import copy

import pytest

from monty.io import zopen
from monty.serialization import loadfn

from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.molecules.atomic import PartialChargesDoc, PartialSpinsDoc
from emmet.core.molecules.bonds import BondingDoc
from emmet.core.molecules.orbitals import OrbitalDoc
from emmet.core.molecules.redox import RedoxDoc
from emmet.core.molecules.thermo import ThermoDoc
from emmet.core.molecules.vibration import VibrationDoc
from emmet.core.molecules.summary import SummaryDoc


@pytest.fixture(scope="session")
def mols(test_dir):

    mols = [MoleculeDoc(**x) for x in loadfn(test_dir / "builder_mols_set.json")]

    return mols


@pytest.fixture(scope="session")
def charges(test_dir):

    charges = [
        PartialChargesDoc(**x) for x in loadfn(test_dir / "builder_charges_set.json")
    ]

    return charges


@pytest.fixture(scope="session")
def spins(test_dir):

    spins = [PartialSpinsDoc(**x) for x in loadfn(test_dir / "builder_spins_set.json")]

    return spins


@pytest.fixture(scope="session")
def bonds(test_dir):

    bonds = [BondingDoc(**x) for x in loadfn(test_dir / "builder_bonding_set.json")]

    return bonds


@pytest.fixture(scope="session")
def orbitals(test_dir):

    orbitals = [OrbitalDoc(**x) for x in loadfn(test_dir / "builder_orbitals_set.json")]

    return orbitals


@pytest.fixture(scope="session")
def redox(test_dir):

    redox = [RedoxDoc(**x) for x in loadfn(test_dir / "builder_redox_set.json")]

    return redox


@pytest.fixture(scope="session")
def thermo(test_dir):

    thermo = [ThermoDoc(**x) for x in loadfn(test_dir / "builder_thermo_set.json")]

    return thermo


@pytest.fixture(scope="session")
def vibes(test_dir):

    vibes = [VibrationDoc(**x) for x in loadfn(test_dir / "builder_vibes_set.json")]

    return vibes

@pytest.mark.skip(reason="Waiting on molecule update.")
def test_summary_doc(mols, charges, spins, bonds, orbitals, redox, thermo, vibes):
    desired_id = "libe-120473"
    docs = {
        "molecules": [e.dict() for e in mols if str(e.molecule_id) == desired_id],
        "partial_charges": [
            e.dict() for e in charges if str(e.molecule_id) == desired_id
        ],
        "partial_spins": [e.dict() for e in spins if str(e.molecule_id) == desired_id],
        "bonding": [e.dict() for e in bonds if str(e.molecule_id) == desired_id],
        "orbitals": [e.dict() for e in orbitals if str(e.molecule_id) == desired_id],
        "redox": [e.dict() for e in redox if str(e.molecule_id) == desired_id],
        "thermo": [e.dict() for e in thermo if str(e.molecule_id) == desired_id],
        "vibration": [e.dict() for e in vibes if str(e.molecule_id) == desired_id],
    }

    for k, v in docs.items():
        if len(v) == 0:
            del docs[k]
        elif len(v) == 1:
            docs[k] = v[0]

    summary_doc = SummaryDoc.from_docs(molecule_id=desired_id, **docs)

    assert summary_doc.property_name == "summary"
    assert summary_doc.electronic_energy is not None
    assert summary_doc.total_enthalpy is not None
    assert summary_doc.frequencies is not None
    assert summary_doc.open_shell == True
    assert summary_doc.nbo_population is not None
    assert summary_doc.bonds is not None
    assert summary_doc.electron_affinity is not None
    print(summary_doc.has_props)



from enum import Enum
from typing import Dict, List, Tuple, TypeVar, Union

from pydantic import Field
from pymatgen.core.structure import Molecule
from pymatgen.analysis.graphs import MoleculeGraph

from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPID
from emmet.core.molecules.orbitals import NaturalPopulation, LonePair, Bond, Interaction


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


T = TypeVar("T", bound="SummaryDoc")


class HasProps(Enum):
    """
    Enum of possible hasprops values.
    """

    molecules = "molecules"
    bonding = "bonding"
    orbitals = "orbitals"
    partial_charges = "partial_charges"
    partial_spins = "partial_spins"
    redox = "redox"
    thermo = "thermo"
    vibration = "vibration"


class SummaryDoc(PropertyDoc):
    """
    Summary information about molecules and their properties, useful for searching.
    """

    property_name = "summary"

    # molecules
    molecule: Molecule = Field(
        ..., description="The lowest energy structure for this molecule"
    )

    task_ids: List[MPID] = Field(
        [],
        title="Calculation IDs",
        description="List of Calculation IDs associated with this molecule.",
    )

    similar_molecules: List[MPID] = Field(
        [], description="IDs associated with similar molecules"
    )

    # thermo
    electronic_energy: float = Field(
        None, description="Electronic energy of the molecule (units: eV)"
    )

    zero_point_energy: float = Field(
        None, description="Zero-point energy of the molecule (units: eV)"
    )

    rt: float = Field(
        None,
        description="R*T, where R is the gas constant and T is temperature, taken "
        "to be 298.15K (units: eV)",
    )

    total_enthalpy: float = Field(
        None, description="Total enthalpy of the molecule at 298.15K (units: eV)"
    )
    total_entropy: float = Field(
        None, description="Total entropy of the molecule at 298.15K (units: eV/K)"
    )

    translational_enthalpy: float = Field(
        None,
        description="Translational enthalpy of the molecule at 298.15K (units: eV)",
    )
    translational_entropy: float = Field(
        None,
        description="Translational entropy of the molecule at 298.15K (units: eV/K)",
    )
    rotational_enthalpy: float = Field(
        None, description="Rotational enthalpy of the molecule at 298.15K (units: eV)"
    )
    rotational_entropy: float = Field(
        None, description="Rotational entropy of the molecule at 298.15K (units: eV/K)"
    )
    vibrational_enthalpy: float = Field(
        None, description="Vibrational enthalpy of the molecule at 298.15K (units: eV)"
    )
    vibrational_entropy: float = Field(
        None, description="Vibrational entropy of the molecule at 298.15K (units: eV/K)"
    )

    free_energy: float = Field(
        None, description="Gibbs free energy of the molecule at 298.15K (units: eV)"
    )

    # vibrational properties
    frequencies: List[float] = Field(
        None, description="List of molecular vibrational frequencies"
    )

    frequency_modes: List[List[List[float]]] = Field(
        None,
        description="Vibrational frequency modes of the molecule (units: Angstrom)",
    )

    ir_intensities: List[float] = Field(
        None,
        title="IR intensities",
        description="Intensities for infrared vibrational spectrum peaks",
    )

    ir_activities: List = Field(
        None,
        title="IR activities",
        description="List indicating if frequency-modes are IR-active",
    )

    # natural bonding orbitals
    open_shell: bool = Field(
        None, description="Is this molecule open-shell (spin multiplicity != 1)?"
    )

    nbo_population: List[NaturalPopulation] = Field(
        None, description="Natural electron populations of the molecule"
    )
    nbo_lone_pairs: List[LonePair] = Field(
        None, description="Lone pair orbitals of a closed-shell molecule"
    )
    nbo_bonds: List[Bond] = Field(
        None, description="Bond-like orbitals of a closed-shell molecule"
    )
    nbo_interactions: List[Interaction] = Field(
        None, description="Orbital-orbital interactions of a closed-shell molecule"
    )

    alpha_population: List[NaturalPopulation] = Field(
        None,
        description="Natural electron populations of the alpha electrons of an "
        "open-shell molecule",
    )
    beta_population: List[NaturalPopulation] = Field(
        None,
        description="Natural electron populations of the beta electrons of an "
        "open-shell molecule",
    )
    alpha_lone_pairs: List[LonePair] = Field(
        None, description="Alpha electron lone pair orbitals of an open-shell molecule"
    )
    beta_lone_pairs: List[LonePair] = Field(
        None, description="Beta electron lone pair orbitals of an open-shell molecule"
    )
    alpha_bonds: List[Bond] = Field(
        None, description="Alpha electron bond-like orbitals of an open-shell molecule"
    )
    beta_bonds: List[Bond] = Field(
        None, description="Beta electron bond-like orbitals of an open-shell molecule"
    )
    alpha_interactions: List[Interaction] = Field(
        None,
        description="Alpha electron orbital-orbital interactions of an open-shell molecule",
    )
    beta_interactions: List[Interaction] = Field(
        None,
        description="Beta electron orbital-orbital interactions of an open-shell molecule",
    )

    # partial charges
    partial_charges: Dict[str, List[float]] = Field(
        None,
        description="Atomic partial charges for the molecule using different partitioning schemes "
        "(Mulliken, Restrained Electrostatic Potential, Natural Bonding Orbitals, etc.)",
    )

    # partial spins
    partial_spins: Dict[str, List[float]] = Field(
        None,
        description="Atomic partial spins for the molecule using different partitioning schemes "
        "(Mulliken, Natural Bonding Orbitals, etc.)",
    )

    # bonding
    molecule_graph: Dict[str, MoleculeGraph] = Field(
        None,
        description="Molecular graph representations of the molecule using different "
        "definitions of bonding.",
    )

    bond_types: Dict[str, Dict[str, List[float]]] = Field(
        description="Dictionaries of bond types to their length under different "
        "definitions of bonding, e.g. C-O to a list of the lengths of "
        "C-O bonds in Angstrom."
    )

    bonds: Dict[str, List[Tuple[int, int]]] = Field(
        description="List of bonds under different definitions of bonding. Each bond takes "
        "the form (a, b), where a and b are 0-indexed atom indices",
    )

    bonds_nometal: Dict[str, List[Tuple[int, int]]] = Field(
        description="List of bonds under different definitions of bonding with all metal ions "
        "removed. Each bond takes the form in the form (a, b), where a and b are "
        "0-indexed atom indices.",
    )

    # redox properties
    electron_affinity: float = Field(
        None, description="Vertical electron affinity in eV"
    )

    ea_id: MPID = Field(None, description="Molecule ID for electron affinity")

    ionization_energy: float = Field(
        None, description="Vertical ionization energy in eV"
    )

    ie_id: MPID = Field(None, description="Molecule ID for ionization energy")

    reduction_free_energy: float = Field(
        None, description="Adiabatic free energy of reduction"
    )

    red_id: MPID = Field(None, description="Molecule ID for adiabatic reduction")

    oxidation_free_energy: float = Field(
        None, description="Adiabatic free energy of oxidation"
    )

    ox_id: MPID = Field(None, description="Molecule ID for adiabatic oxidation")

    reduction_potentials: Dict[str, float] = Field(
        None, description="Reduction potentials with various " "reference electrodes"
    )

    oxidation_potentials: Dict[str, float] = Field(
        None, description="Oxidation potentials with various " "reference electrodes"
    )

    # has props
    has_props: List[HasProps] = Field(
        None, description="List of properties that are available for a given material."
    )

    @classmethod
    def from_docs(cls, molecule_id: MPID, **docs: Dict[str, Union[Dict, List[Dict]]]):
        """Converts a bunch of property docs into a SummaryDoc"""

        doc = _copy_from_doc(docs)
        doc["has_props"] = list(set(doc["has_props"]))

        return SummaryDoc(molecule_id=molecule_id, **doc)


# Key mapping
summary_fields: Dict[str, list] = {
    HasProps.molecules.value: [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "composition",
        "formula_alphabetical",
        "chemsys",
        "symmetry",
        "molecule",
        "deprecated",
        "task_ids",
    ],
    HasProps.thermo.value: [
        "electronic_energy",
        "zero_point_energy",
        "rt",
        "total_enthalpy",
        "total_entropy",
        "translational_enthalpy",
        "translational_entropy",
        "rotational_enthalpy",
        "rotational_entropy",
        "vibrational_enthalpy",
        "vibrational_entropy",
        "free_energy",
    ],
    HasProps.vibration.value: [
        "frequencies",
        "frequency_modes",
        "ir_intensities",
        "ir_activities",
    ],
    HasProps.orbitals.value: [
        "open_shell",
        "nbo_population",
        "nbo_lone_pairs",
        "nbo_bonds",
        "nbo_interactions",
        "alpha_population",
        "beta_population",
        "alpha_lone_pairs",
        "beta_lone_pairs",
        "alpha_bonds",
        "beta_bonds",
        "alpha_interactions",
        "beta_interactions",
    ],
    HasProps.partial_charges.value: ["partial_charges"],
    HasProps.partial_spins.value: ["partial_spins"],
    HasProps.bonding.value: ["molecule_graph", "bond_types", "bonds", "bonds_nometal"],
    HasProps.redox.value: [
        "electron_affinity",
        "ea_id",
        "ionization_energy",
        "ie_id",
        "reduction_free_energy",
        "red_id",
        "oxidation_free_energy",
        "ox_id",
        "reduction_potentials",
        "oxidation_potentials",
    ],
}


def _copy_from_doc(doc):
    """Helper function to copy the list of keys over from amalgamated document"""

    d = {"has_props": []}

    # Function to grab the keys and put them in the root doc
    for doc_key in summary_fields:
        sub_doc = doc.get(doc_key, None)
        if isinstance(sub_doc, list) and len(sub_doc) > 0:
            d["has_props"].append(doc_key)
            for copy_key in summary_fields[doc_key]:
                d[copy_key] = dict()
                for sub_item in sub_doc:
                    # In cases where multiple docs have the same properties,
                    # they must differ by method
                    if copy_key in sub_item and "method" in sub_item:
                        d[copy_key][sub_item["method"]] = sub_item[copy_key]

        elif isinstance(sub_doc, dict):
            d["has_props"].append(doc_key)
            d.update(
                {
                    copy_key: sub_doc[copy_key]
                    for copy_key in summary_fields[doc_key]
                    if copy_key in sub_doc
                }
            )

    return d



import json
import datetime

import pytest

from monty.io import zopen
from monty.serialization import loadfn

from emmet.core.qchem.calc_types import TaskType
from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.qchem.task import TaskDocument


@pytest.fixture(scope="session")
def test_tasks(test_dir):
    with zopen(test_dir / "liec_tasks.json.gz") as f:
        data = json.load(f)

    for d in data:
        d["last_updated"] = datetime.datetime.strptime(
            d["last_updated"]["string"], "%Y-%m-%d %H:%M:%S.%f"
        )

    tasks = [TaskDocument(**t) for t in data]
    return tasks


@pytest.mark.xfail
def test_make_mol(test_tasks):

    molecule = MoleculeDoc.from_tasks(test_tasks)
    assert molecule.formula_alphabetical == "C3 H4 Li1 O3"
    assert len(molecule.task_ids) == 5
    assert len(molecule.entries) == 5

    bad_task_group = [
        task
        for task in test_tasks
        if task.task_type
        not in [
            TaskType.Geometry_Optimization,
            TaskType.Frequency_Flattening_Geometry_Optimization,
        ]
    ]

    with pytest.raises(Exception):
        MoleculeDoc.from_tasks(bad_task_group)


def test_make_deprecated_mol(test_tasks):
    bad_task_group = [
        task
        for task in test_tasks
        if task.task_type
        not in [
            TaskType.Geometry_Optimization,
            TaskType.Frequency_Flattening_Geometry_Optimization,
        ]
    ]

    molecule = MoleculeDoc.construct_deprecated_molecule(bad_task_group)

    assert molecule.deprecated
    assert molecule.formula_alphabetical == "C3 H4 Li1 O3"
    assert len(molecule.task_ids) == 4
    assert molecule.entries is None


def test_schema():
    MoleculeDoc.schema()



import pytest
from pymatgen.core import Structure
from monty.serialization import loadfn

from emmet.core.magnetism import MagnetismDoc
from emmet.core.utils import jsanitize


@pytest.fixture(scope="session")
def magnetism_test_data(test_dir):
    return loadfn(test_dir / "magnetism/magnetism_mats_sample.json")


def test_magnetism_doc(magnetism_test_data):
    test_orderings = {"mp-1034331": "FM", "mp-753472": "NM"}

    for material in magnetism_test_data:
        structure = Structure.from_dict(jsanitize(material["structure"]))
        total_magnetization = material["magnetism"]["total_magnetization"]
        material_id = material["task_id"]

        doc = MagnetismDoc.from_structure(
            structure=structure,
            material_id=material_id,
            total_magnetization=total_magnetization,
            deprecated=False,
        )
        assert doc is not None
        assert doc.total_magnetization == pytest.approx(abs(total_magnetization))

        if material_id in test_orderings:
            assert doc.ordering == test_orderings[material_id]



""" Core definition of Structure metadata """
from __future__ import annotations

from typing import List, Optional, Type, TypeVar

from pydantic import Field
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure, Molecule

from emmet.core.base import EmmetBaseModel
from emmet.core.symmetry import SymmetryData, PointGroupData

T = TypeVar("T", bound="StructureMetadata")
S = TypeVar("S", bound="MoleculeMetadata")


class StructureMetadata(EmmetBaseModel):
    """
    Mix-in class for structure metadata
    """

    # Structure metadata
    nsites: int = Field(None, description="Total number of sites in the structure.")
    elements: List[Element] = Field(None, description="List of elements in the material.")
    nelements: int = Field(None, description="Number of elements.")
    composition: Composition = Field(None, description="Full composition for the material.")
    composition_reduced: Composition = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition.",
    )
    formula_pretty: str = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula.",
    )
    formula_anonymous: str = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula.",
    )
    chemsys: str = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the material.",
    )
    volume: float = Field(
        None,
        title="Volume",
        description="Total volume for this structure in Angstroms^3.",
    )

    density: float = Field(None, title="Density", description="Density in grams per cm^3.")

    density_atomic: float = Field(
        None,
        title="Packing Density",
        description="The atomic packing density in atoms per cm^3.",
    )

    symmetry: SymmetryData = Field(None, description="Symmetry data for this material.")

    @classmethod
    def from_composition(
        cls: Type[T],
        composition: Composition,
        fields: Optional[List[str]] = None,
        **kwargs,
    ) -> T:

        fields = (
            [
                "elements",
                "nelements",
                "composition",
                "composition_reduced",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
            ]
            if fields is None
            else fields
        )
        composition = composition.remove_charges()

        elsyms = sorted(set([e.symbol for e in composition.elements]))

        data = {
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": composition,
            "composition_reduced": composition.reduced_composition.remove_charges(),
            "formula_pretty": composition.reduced_formula,
            "formula_anonymous": composition.anonymized_formula,
            "chemsys": "-".join(elsyms),
        }

        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)

    @classmethod
    def from_structure(
        cls: Type[T],
        meta_structure: Structure,
        fields: Optional[List[str]] = None,
        **kwargs,
    ) -> T:

        fields = (
            [
                "nsites",
                "elements",
                "nelements",
                "composition",
                "composition_reduced",
                "formula_pretty",
                "formula_anonymous",
                "chemsys",
                "volume",
                "density",
                "density_atomic",
                "symmetry",
            ]
            if fields is None
            else fields
        )
        comp = meta_structure.composition.remove_charges()
        elsyms = sorted(set([e.symbol for e in comp.elements]))
        symmetry = SymmetryData.from_structure(meta_structure)

        data = {
            "nsites": meta_structure.num_sites,
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
            "volume": meta_structure.volume,
            "density": meta_structure.density,
            "density_atomic": meta_structure.volume / meta_structure.num_sites,
            "symmetry": symmetry,
        }
        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)


class MoleculeMetadata(EmmetBaseModel):
    """
    Mix-in class for molecule metadata
    """

    charge: float = Field(None, description="Charge of the molecule")
    spin_multiplicity: int = Field(None, description="Spin multiplicity of the molecule")
    natoms: int = Field(None, description="Total number of atoms in the molecule")
    elements: List[Element] = Field(None, description="List of elements in the molecule")
    nelements: int = Field(None, title="Number of Elements")
    composition: Composition = Field(None, description="Full composition for the molecule")
    formula_alphabetical: str = Field(
        None,
        title="Alphabetical Formula",
        description="Alphabetical molecular formula",
    )
    chemsys: str = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the molecule",
    )

    symmetry: PointGroupData = Field(None, description="Symmetry data for this molecule")

    @classmethod
    def from_molecule(
        cls: Type[S], meta_molecule: Molecule, fields: Optional[List[str]] = None, **kwargs
    ) -> MoleculeMetadata:

        fields = (
            [
                "charge",
                "spin_multiplicity",
                "natoms",
                "elements",
                "nelements",
                "composition",
                "formula_alphabetical",
                "chemsys",
                "symmetry",
            ]
            if fields is None
            else fields
        )
        comp = meta_molecule.composition
        elsyms = sorted(set([e.symbol for e in comp.elements]))
        symmetry = PointGroupData.from_molecule(meta_molecule)

        data = {
            "charge": meta_molecule.charge,
            "spin_multiplicity": meta_molecule.spin_multiplicity,
            "nsites": len(meta_molecule),
            "elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "formula_alphabetical": comp.alphabetical_formula,
            "chemsys": "-".join(elsyms),
            "symmetry": symmetry,
        }

        return cls(**{k: v for k, v in data.items() if k in fields}, **kwargs)
